#include <Message.h>
#include <TextControl.h>
#include <Application.h>
#include <MenuField.h>
#include <Menu.h>
#include <MenuItem.h>

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "ViewUtils.h"

#include "ConfigView.h"

#define INTERFACE_ELEMENTS_SUCK 1

#if INTERFACE_ELEMENTS_SUCK
void align_text_controls(BView* view);
#endif

/*----------------------------------------------------------------------------------------
	° MountainConfigView
------------------------------------------------------------------------------------------*/

BArchivable* MountainConfigView::Instantiate(BMessage* archive)
{	
	if(validate_instantiation(archive,"MountainConfigView"))
		return new MountainConfigView(archive);
	return NULL;

}

status_t MountainConfigView::Archive(BMessage* archive,bool deep) const
{	
	BView::Archive(archive,deep);

	archive->AddString("class","MountainConfigView");

	return B_OK;
}

MountainConfigView::MountainConfigView(BMessage* archive) : BView(archive)
{
	mRenderer = new MountainRenderer;
	SyncControls();

	mQuitApp = true;
}


MountainConfigView::~MountainConfigView()
{
	delete mRenderer;
}

/*----------------------------------------------------------------------------------------
	° MessageReceived
------------------------------------------------------------------------------------------*/

void MountainConfigView::MessageReceived(BMessage* msg)
{
	BControl* cntl;
	msg->FindPointer("source", (void**)&cntl);
	
	if(cntl == NULL) return;
	
	BTextControl* txtCntl = cast_as(cntl,BTextControl);
	const char* text = NULL;
	if(txtCntl != NULL)
		text = const_cast<char*>(txtCntl->Text());

	switch(msg->what)
	{
// Perspective
		case 'Pspc':
			mRenderer->map = cntl->Value();
		break;
// Map
		case 'Map ':
			mRenderer->map = !cntl->Value();
		break;
// Width:
		case 'Wdth':
			mRenderer->s_width = atoi(text);
		break;		
// Height
		case 'Hght':
			mRenderer->s_height = atoi(text);
		break;
// Vertical Stretch		
		case 'Strc':
			mRenderer->stretch = atof(text);
		break;
// Vertical Shift
		case 'Shft':
			mRenderer->pre_shift = atof(text);
		break;		
// Vertical Illumination angle
		case 'VIlm':
			mRenderer->phi = M_PI * atof(text) / 180.0;
			if ( mRenderer->phi  < 0.0 )
				mRenderer->phi =0.0;
			else if( mRenderer->phi > M_PI/2.0 )
				mRenderer->phi  = M_PI/2.0;
		break;
// Horizontal Illumination angle
		case 'HIlm':
			mRenderer->alpha = M_PI * atof(text) / 180.0;
			if( mRenderer->alpha < -M_PI/3.0 )
				mRenderer->alpha = -M_PI/3.0;
			else if( mRenderer->alpha > M_PI/3.0 )
				mRenderer->alpha = M_PI/3.0;
		break;
//Snooze time
		case 'Zzzz':
			mRenderer->snooze_time = atoi(text);
			if(mRenderer->snooze_time < 0);
				mRenderer->snooze_time = 0;
		break;
// Scroll Columns
		case 'Cols':
			mRenderer->repeat = atoi(text);
		break;
// Viewpoint altitdue
		case 'VAlt':
			mRenderer->altitude = atof(text);
		break;
// Viewpoint distance
		case 'VDst':
			mRenderer->distance = atof(text);
		break;
// Ambient light
		case 'Ambt':
			mRenderer->ambient = atof(text);
			if( mRenderer->ambient < 0.0 )
				mRenderer->ambient = 0.0;
			if( mRenderer->ambient > 1.0 )
				mRenderer->ambient=1.0;
		break;
// Vertical light
		case 'VLgt':
			mRenderer->vfract = atof(text);
			if( mRenderer->vfract < 0.0 )
				mRenderer->vfract = 0.0;
			break;	
// Contrast
		case 'Ctst':
			mRenderer->contrast = atof(text);
			if( mRenderer->contrast < 0.0 )
				mRenderer->contrast=0.0;
		break;
// Use Water Reflection
		case 'Rfct':
			mRenderer->reflec = cntl->Value();
		break;
// # of colors
		case 'Colr':
			mRenderer->n_col= atoi(text);
		break;
// water altitude
		case 'Aqua':
			mRenderer->sealevel = atof(text);
		break;
// mean
		case 'Mean':
			mRenderer->forceheight = atof(text);
		break;
// variation
		case 'Var ':
			mRenderer->slope = atoi(text);
		break;
// Contour
		case 'Cont':
			mRenderer->contour = atof(text);
		break;
// Algorithm Stuff:
		case 'Dim ':
			mRenderer->f_dim = atof(text);
			if( mRenderer->f_dim < 0.5 )
			 	mRenderer->f_dim=0.5;
			else if( mRenderer->f_dim > 1.0 )
			 	mRenderer->f_dim=1.0;
		break;
		
		case 'Seed':
			mRenderer->seed = atoi(text);
		break;
		
		case 'Iter':
			mRenderer->stop = atoi(text);
			if(mRenderer->stop < 0)
				mRenderer->stop = 0;
		break;

		case 'Step':
			mRenderer->smooth = atoi(text);
		break;

		case '-X  ':
			mRenderer->mix = atof(text);
		break;
		
		case '-Y  ':
			mRenderer->midmix = atof(text);
		break;
		
		case 'Cros':
			mRenderer->cross = cntl->Value();
		break;
		
		case 'Cncl':
			if(mQuitApp)
				be_app->PostMessage(B_QUIT_REQUESTED);
			else
				Window()->Close();
		break;
		
		default:
			BView::MessageReceived(msg);
		break;
	}
}

/*----------------------------------------------------------------------------------------
	° AttachedToWindow
------------------------------------------------------------------------------------------*/

void MountainConfigView::AttachedToWindow()
{
	ViewUtils::SetAllSubControlsTarget(this, this);
	((BMenuField*)FindView("Display Type"))->Menu()->SetTargetForItems(this);
	align_text_controls(this);
}

/*----------------------------------------------------------------------------------------
	° SyncControls
------------------------------------------------------------------------------------------*/

void MountainConfigView::SyncControls(const MountainRenderer* r)
{
	if(r != NULL)
		*mRenderer = *r;
		
	((BMenuField*)FindView("Display Type"))->Menu()->ItemAt(((mRenderer->map)?1:0))->SetMarked(true);

	BControl* cntl = NULL;
	BTextControl* txtCtl = NULL;
	char txt[32];
	
	sprintf(txt, "%d", mRenderer->s_width);
	txtCtl = cast_as(FindView("Width"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);
		
	sprintf(txt, "%d", mRenderer->s_height);
	txtCtl = cast_as(FindView("Height"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->stretch);
	txtCtl = cast_as(FindView("Stretch"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->pre_shift);
	txtCtl = cast_as(FindView("Shift"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->phi * 180.0 / M_PI);
	txtCtl = cast_as(FindView("Vertical"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->alpha * 180.0 / M_PI);
	txtCtl = cast_as(FindView("Horizontal"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%d", mRenderer->snooze_time);
	txtCtl = cast_as(FindView("Snooze"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%d", mRenderer->repeat);
	txtCtl = cast_as(FindView("Columns"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->altitude);
	txtCtl = cast_as(FindView("Altitude"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->distance);
	txtCtl = cast_as(FindView("Distance"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->ambient);
	txtCtl = cast_as(FindView("Ambient"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->vfract);
	txtCtl = cast_as(FindView("Vertical Light"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->contrast);
	txtCtl = cast_as(FindView("Contrast"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	cntl = cast_as(FindView("Water Reflections"),BControl);
	if(cntl) cntl->SetValue(mRenderer->reflec);

	sprintf(txt, "%d", mRenderer->n_col);
	txtCtl = cast_as(FindView("Colors"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->sealevel);
	txtCtl = cast_as(FindView("Water"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->forceheight);
	txtCtl = cast_as(FindView("Mean"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%d", mRenderer->slope);
	txtCtl = cast_as(FindView("Variation"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->contour);
	txtCtl = cast_as(FindView("Contour"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->f_dim);
	txtCtl = cast_as(FindView("Dimension"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%d", mRenderer->seed);
	txtCtl = cast_as(FindView("Seed"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%d", mRenderer->stop);
	txtCtl = cast_as(FindView("Iterations"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);
		
	sprintf(txt, "%d", mRenderer->smooth);
	txtCtl = cast_as(FindView("Steps"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);
	
	sprintf(txt, "%.3f", mRenderer->mix);
	txtCtl = cast_as(FindView("-X"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);

	sprintf(txt, "%.3f", mRenderer->mix);
	txtCtl = cast_as(FindView("-Y"),BTextControl);
	if(txtCtl) txtCtl->SetText(txt);
		
	cntl = cast_as(FindView("Cross Updates"),BControl);
	if(cntl) cntl->SetValue(mRenderer->cross);
}
#if INTERFACE_ELEMENTS_SUCK
/*----------------------------------------------------------------------------------------
	° align_text_controls
------------------------------------------------------------------------------------------*/
// get around IE deficiency, so what else is new?
void align_text_controls(BView* view)
{
	for(int i = 0; BView* subView = view->ChildAt(i); i++)
	{
		BTextControl* cntl = cast_as(subView, BTextControl);
		if(cntl)
			cntl->SetAlignment(B_ALIGN_RIGHT,B_ALIGN_LEFT);
		align_text_controls(subView);
	}
}
#endif