#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "MountainRenderer.h"

#define SIDE 1.0  

extern void seed_uni(int);

/*----------------------------------------------------------------------------------------
	° MountainRenderer
------------------------------------------------------------------------------------------*/

MountainRenderer::MountainRenderer()
{	
	mCurCol = 0;
	mScrolling = false;
	mImage = NULL;

	pre_shift = .5;

	Fold *top;
	levels = 10;
	stop = 2;
	smooth = 1;
	cross = TRUE;
	slope = 1;
	snooze_time = 0;
	n_col=DEF_COL;
	band_size=BAND_SIZE;
	request_clear = FALSE;
	f_dim = 0.65;
	mix   =0.0;
	midmix=0.0;
	start;      /* starting value for the surface */
	mean=0.0;   /* mean value of surface */
	varience;   /* rough estimate of the s_height of the range */
	shift=0.5;    /* offset from calcalt to artist coordinates */
	delta_shadow; /* offset of shadow at each step */
	stretch=0.6;   /* vertical stretch */
	contour = 0.3;
	ambient = 0.3;  /* level of ambient light */
	contrast = 1.0; /* contrast,
	                       * increases or decreases effect of cosine rule */
	vfract   = 0.6; /* relative strength of vertical light relative
	                        * to the main light source
	                        */
	altitude = 2.5;
	distance = 4.0;
	phi=(40.0 * M_PI)/180.0; /* angle of the light (vertical plane)*/
	alpha=0.0;             /* angle of the light (horizontal plane)
	                               * must have -pi/4 < alpha < pi/4
	                               */
	shadow_slip;
	shadow_register=0.0;
	cos_phi;
	sin_phi;
	tan_phi;
	x_fact;
	y_fact;
	sealevel = 0.0;
	forceheight  = -1.0;
	width;        /* width of the landscape, (function of levels) */
	seed=0;       /* zero means read the clock */
	
	/*
	 * viewport parameters
	 */
	vangle;                    /* view angle 0 == horizontal
	                                   *increase to look down
	                                   */
	tan_vangle;                                     
	vscale;           /* rescale physical s_height by this amount. */
	viewpos;        /* position of viewpoint */
	viewheight;      /* s_height of viewpoint */
	focal;                  /* focal length, calc to preserve aspect ratio */
	
	Height *shadow;               /* s_height of the shadows */
	Height *a_strip, *b_strip;    /* the two most recent strips */

	plot_saved = false;
	last_plot_x = 0;
	//table = NULL;
	
	base=0;
	
	repeat=1;
 	map = 0;
 	reflec = 0;

	seed_uni(seed);
	
	s_height = 512;
	s_width  = 512;
}

/*----------------------------------------------------------------------------------------
	° PlotNextColumn
------------------------------------------------------------------------------------------*/

void MountainRenderer::PlotNextColumn(BBitmap* image)
{
	mImage = image;
	
	if( s_height > s_width )
	{
		mapwid=s_width;
	}else{
		mapwid=s_height;
	}
	if(!mScrolling)
	{
		if( repeat >= 0 )
		{
			plot_column(mCurCol++,map,reflec,0);	 	
			mScrolling = (mCurCol >= s_width);
			if(mScrolling)
				mCurCol = s_width - repeat - 1;
		}else
		{
			plot_column(mCurCol--,map,reflec,0);
			mScrolling = (mCurCol < 0);
			if(mScrolling)
				mCurCol = -1 - repeat;
		}
	}else
	{
		// do the scroll 
		if(repeat > 0)
		{
			plot_column(mCurCol++,map,reflec,0);
			
			if(mCurCol >= (s_width-1))
			{
				scroll_screen();
				mCurCol = s_width - repeat - 1 ;
			}
		}else
		{
			plot_column(mCurCol--,map,reflec,0);
			if(mCurCol < 0)
			{
				scroll_screen();
				mCurCol = -1 - repeat;
			}
		}
	}

}
// WARNING! WARNING! PORTED CODE FOLLOWS WARNING! WARNING! //
/*----------------------------------------------------------------------------------------
	° plot_pixel
------------------------------------------------------------------------------------------*/

void MountainRenderer::plot_pixel(int x,int y,Gun value)
{
  int do_draw, draw_x, draw_y1, draw_y2;
  rgb_color draw_colour;
  
  // if x is negative this means flush the stored request 
  if(! plot_saved) // no stored values 
  {
    plot_x = x;
    plot_y1=plot_y2 = y;
    plot_col=table[value];
    do_draw=FALSE;
    plot_saved = (x >=0);
  }else{
    if( x < 0 )  // requesting a flush 
    {
      draw_x=plot_x;
      draw_y1=plot_y1;
      draw_y2=plot_y2;
      draw_colour = plot_col;
      plot_saved=FALSE;
      do_draw=TRUE;
    }else{      // plot request with saved value 
      if( (x==plot_x) && 
      	(plot_col.red == table[value].red) &&
      	(plot_col.blue == table[value].blue) &&
      	(plot_col.green == table[value].green)	) // add to line 
      {
        if(y<plot_y1) plot_y1=y;
        if(y>plot_y2) plot_y2=y;
        do_draw=FALSE;
      }else{
        draw_x=plot_x;
        draw_y1=plot_y1;
        draw_y2=plot_y2;
        draw_colour=plot_col;
        do_draw=TRUE;
        plot_x=x;
        plot_y1=plot_y2=y;
        plot_col=table[value];
      }
    }
  }
	if( do_draw)
	{
		mImage->Lock();
		
		BView* view = mImage->ChildAt(0);
		if(view)
		{
			view->SetHighColor(draw_colour);
			BPoint start(draw_x,draw_y1);
		
			if( draw_y1 == draw_y2 ){
				view->StrokeLine(start,start);
			}else
			{
				BPoint	end(draw_x,draw_y2);
				view->StrokeLine(start,end);
			}
		}
		//view->Sync();
		mImage->Unlock();
	}
}


/*----------------------------------------------------------------------------------------
	° next_col
------------------------------------------------------------------------------------------*/

Col *MountainRenderer::next_col ( int paint, int reflec)
{
  Col *res;
  int i,offset=0;
  
  /*{{{  update strips */
  if(paint)
  {
    if(reflec)
    {
      res = mirror( a_strip,b_strip,shadow);
    }else{
      res = camera( a_strip,b_strip,shadow);
    }
  }else{
    res = makemap(a_strip,b_strip,shadow);
  }
  free(a_strip);
  a_strip=b_strip;
  b_strip = extract( next_strip(top) );
  /*}}}*/

  /*{{{update the shadows*/
  /* shadow_slip is the Y component of the light vector.
   * The shadows can only step an integer number of points in the Y
   * direction so we maintain shadow_register as the deviation between
   * where the shadows are and where they should be. When the magnitude of
   * this gets larger then 1 the shadows are slipped by the required number of
   * points.
   * This will not work for very oblique angles so the horizontal angle
   * of illumination should be constrained.
   */
  shadow_register += shadow_slip;
  if( shadow_register >= 1.0 )
  {
    /*{{{negative offset*/
    while( shadow_register >= 1.0 )
    {
      shadow_register -= 1.0;
      offset++;
    }
    for(i=width-1 ; i>=offset ; i--)
    {
      shadow[i] = shadow[i-offset]-delta_shadow;
      if( shadow[i] < b_strip[i] )
      {
        shadow[i] = b_strip[i];
      }
      /*{{{  stop shadow at sea level */
      if( shadow[i] < sealevel )
      {
        shadow[i] = sealevel;
      }
      /*}}}*/
    }
    for(i=0;i<offset;i++)
    {
      shadow[i] = b_strip[i];
      /*{{{  stop shadow at sea level*/
      if( shadow[i] < sealevel )
      {
        shadow[i] = sealevel;
      }
      /*}}}*/
    }
    /*}}}*/
  }else if( shadow_register <= -1.0 ){
    /*{{{positive offset*/
    while( shadow_register <= -1.0 )
    {
      shadow_register += 1.0;
      offset++;
    }
    for(i=0 ; i<width-offset ; i++)
    {
      shadow[i] = shadow[i+offset]-delta_shadow;
      if( shadow[i] < b_strip[i] )
      {
        shadow[i] = b_strip[i];
      }
      /*{{{  stop shadow at sea level */
      if( shadow[i] < sealevel )
      {
        shadow[i] = sealevel;
      }
      /*}}}*/
    }
    for(;i<width;i++)
    {
      shadow[i] = b_strip[i];
      /*{{{  stop shadow at sea level*/
      if( shadow[i] < sealevel )
      {
        shadow[i] = sealevel;
      }
      /*}}}*/
    }
    /*}}}*/
  }else{
    /*{{{no offset*/
    for(i=0 ; i<width ; i++)
    {
      shadow[i] -= delta_shadow;
      if( shadow[i] < b_strip[i] )
      {
        shadow[i] = b_strip[i];
      }
      /*{{{  stop shadow at sea level */
      if( shadow[i] < sealevel )
      {
        shadow[i] = sealevel;
      }
      /*}}}*/
    }
    /*}}}*/
  }
  /*}}}*/
  
  return(res);
}


/*----------------------------------------------------------------------------------------
	° plot_column
------------------------------------------------------------------------------------------*/

void MountainRenderer::plot_column(int p,int map,int reflec,int snooze)
{
  Col *l;
  int j;
  
  l = next_col(1-map,reflec); 
  if( map )
  {
    for( j=0 ;j<(s_height-mapwid); j++)
    {
      plot_pixel(p,((s_height-1)-j),BLACK);
    }
    for(j=0; j<mapwid ; j++)
    {
      plot_pixel(p,((mapwid-1)-j),l[j]);
    }
  }else{
    for(j=0 ; j<s_height ; j++)
    {
      /* we assume that the scroll routine fills the
       * new region with a SKY value. This allows us to
       * use a testured sky for B/W displays
       */
      if( l[j] != SKY )
      {
        plot_pixel(p,((s_height-1)-j),l[j]);
      }
    }
  }
  free(l);
// flush_region(0,1,p,0);
// zap_events(snooze);
}

/*----------------------------------------------------------------------------------------
	° scroll_screen
------------------------------------------------------------------------------------------*/

void MountainRenderer::scroll_screen()
{
	mImage->Lock();
	BView* view = mImage->ChildAt(0);
	
	if(view)
	{	
		BRect src = view->Bounds();
		BRect dst = src;
		dst.OffsetBy(-repeat,0);
		view->CopyBits(src,dst);
			
		BRect r;
		r.top = 0; r.bottom = s_height;
		if(repeat > 0)
		{	
			r.left = s_width - repeat;
			r.right= s_width;
		}else
		{
			r.left = 0;
			r.right= -repeat - 1;
		}
		rgb_color color;
		color.red = color.green = color.blue = 0xFF;
		view->SetHighColor(color);
		view->FillRect(r);
	}
	mImage->Unlock();
}

#pragma mark -
#pragma mark === init ===
/*----------------------------------------------------------------------------------------
	° set_clut
------------------------------------------------------------------------------------------*/

void MountainRenderer::set_clut (int max_col,Gun *red,Gun *green,Gun *blue)
{
  int band,shade;
  float top, bot;
  float intensity;
  int tmp;
/*
*  float rb[N_BANDS] = { 0.167,0.200,0.333,0.450,0.600,1.000 };
*  float gb[N_BANDS] = { 0.667,0.667,0.500,0.500,0.600,1.000 };
*  float bb[N_BANDS] = { 0.500,0.450,0.333,0.200,0.000,1.000 };
*/

  float rb[N_BANDS];
  float gb[N_BANDS];
  float bb[N_BANDS];

  /* band base colours as RGB fractions */
  rb[0] = 0.450; rb[1] = 0.600; rb[2] = 1.000;
  gb[0] = 0.500; gb[1] = 0.600; gb[2] = 1.000;
  bb[0] = 0.333; bb[1] = 0.000; bb[2] = 1.000;

  /*{{{  black */
  red[BLACK]       = 0;
  green[BLACK]     = 0;
  blue[BLACK]      = 0;
  /*}}}*/
  /*{{{  white */
  red[WHITE]       = COL_RANGE;
  green[WHITE]     = COL_RANGE;
  blue[WHITE]      = COL_RANGE;
  /*}}}*/
  /*{{{  sky*/
  red[SKY]         = 0.404*COL_RANGE;
  green[SKY]       = 0.588*COL_RANGE;
  blue[SKY]        = COL_RANGE;
  /*}}}*/
  /*{{{  sea (lit) */
  red[SEA_LIT]     = 0;
  green[SEA_LIT]   = 0.500*COL_RANGE;
  blue[SEA_LIT]    = 0.700*COL_RANGE;
  /*}}}*/
  /*{{{  sea (unlit)*/
  red[SEA_UNLIT]   = 0;
  green[SEA_UNLIT] = ((ambient+(vfract/(1.0+vfract)))*0.500)*COL_RANGE;
  blue[SEA_UNLIT]  = ((ambient+(vfract/(1.0+vfract)))*0.700)*COL_RANGE;
  /*}}}*/

  if( MIN_COL > max_col )
  {
    fprintf(stderr,"set_clut: less than the minimum number of colours available\n");
    exit(1);
  }
  /* max_col can over-rule band_size */
  while( (BAND_BASE + band_size*N_BANDS) > max_col )
  {
    band_size--;
  }
    
  for( band=0 ; band<N_BANDS; band++)
  {
    for(shade=0 ; shade < band_size ; shade++)
    {
      if( (BAND_BASE + (band*band_size) + shade) >= max_col )
      {
        fprintf(stderr,"INTERNAL ERROR, overflowed clut\n");
        exit(1);
      }
      /*{{{  set red */
      top = rb[band];
      bot = ambient * top;
      intensity = bot + ((shade * (top - bot))/(band_size-1));
      tmp = COL_RANGE * intensity;
      if (tmp < 0)
      {
        fprintf(stderr,"set_clut: internal error: invalid code %d\n",tmp);
        exit(2);
      }
      if( tmp > COL_RANGE )
      {
        tmp = COL_RANGE;
      }
      red[BAND_BASE + (band*band_size) + shade] = tmp;
      /*}}}*/
      /*{{{  set green */
      top = gb[band];
      bot = ambient * top;
      intensity = bot + ((shade * (top - bot))/(band_size-1));
      tmp = COL_RANGE * intensity;
      if (tmp < 0)
      {
        fprintf(stderr,"set_clut: internal error: invalid code %d\n",tmp);
        exit(2);
      }
      if( tmp > COL_RANGE )
      {
        tmp = COL_RANGE;
      }
      green[BAND_BASE + (band*band_size) + shade] = tmp;
      /*}}}*/
      /*{{{  set blue */
      top = bb[band];
      bot = ambient * top;
      intensity = bot + ((shade * (top - bot))/(band_size-1));
      tmp = COL_RANGE * intensity;
      if (tmp < 0)
      {
        fprintf(stderr,"set_clut: internal error: invalid code %d\n",tmp);
        exit(2);
      }
      if( tmp > COL_RANGE )
      {
        tmp = COL_RANGE;
      }
      blue[BAND_BASE + (band*band_size) + shade] = tmp;
      /*}}}*/
    }
  }
}

/*----------------------------------------------------------------------------------------
	° init_graphics
------------------------------------------------------------------------------------------*/

void MountainRenderer::init_graphics()
{
	mCurCol = 0;
	shift = pre_shift;
	Gun *clut[3];
	for(int i=0 ;i<3 ;i++)
	{
	 	clut[i] = (Gun *) malloc(n_col * sizeof(Gun));
	}
	set_clut(n_col,clut[0], clut[1], clut[2]);
	
	 for(int i=0; i<n_col ; i++)
	 {
	   table[i].red   = clut[0][i];
	   table[i].green = clut[1][i];
	   table[i].blue  = clut[2][i];
	 }

	for(int i=0;i<3;i++)
	{
	 	free(clut[i]);
	}

	init_artist_variables();
}

/*----------------------------------------------------------------------------------------
	° init_artist_variables
------------------------------------------------------------------------------------------*/

void MountainRenderer::init_artist_variables()
{
  float dh, dd;
  int pwidth;  /* longest lengthscale for update */
  
  width= (1 << levels)+1;
  pwidth= (1 << (levels - stop))+1;

  /* make the fractal SIDE wide, this makes it easy to predict the
   * average s_height returned by calcalt. If we have stop != 0 then
   * make the largest update length = SIDE
   */
  cos_phi = cos( phi );
  sin_phi = sin( phi );
  tan_phi = tan( phi );

  x_fact = cos_phi* cos(alpha);
  y_fact = cos_phi* sin(alpha);
  vscale = stretch * pwidth;  /* have approx same s_height as fractal width
                               * this makes each pixel SIDE=1.0 wide.
                               * c.f. get_col
                               */

  delta_shadow = tan_phi /cos(alpha);
  shadow_slip = tan(alpha);
  /* guess the average s_height of the fractal */
  varience = pow( SIDE ,(2.0 * f_dim));
  varience = vscale * varience ;
  shift = shift * varience;
  varience = varience + shift;

  start = (sealevel - shift) / vscale ; /* always start at sealevel */ 

  /* set the position of the view point */
  viewheight = altitude * width;
  viewpos = - distance * width;

  /* set viewing angle and focal length (vertical-magnification)
   * try mapping the bottom of the fractal to the bottom of the
   * screen. Try to get points in the middle of the fractal
   * to be 1 pixel high
   */
  dh = viewheight;
  dd = (width / 2.0) - viewpos;
  focal = sqrt( (dd*dd) + (dh*dh) );
#ifndef SLOPPY
  tan_vangle = (double) ((double)(viewheight-sealevel)/(double) - viewpos);
  vangle = atan ( tan_vangle );
  vangle -= atan( (double) (s_height/2) / focal ); 
#else
  /* we are making some horrible approximations to avoid trig funtions */
  tan_vangle = (double) ((double)(viewheight-sealevel)/(double) - viewpos);
  tan_vangle = tan_vangle - ( (double) (s_height/2) / focal );
#endif

  fold_param.mean=mean;
  fold_param.rg1=smooth & 4;
  fold_param.rg2=smooth & 2;
  fold_param.rg3=smooth & 1;
  fold_param.cross=cross;
  fold_param.force_front=slope;
  fold_param.force_back=0;
  fold_param.forceval=forceheight;
  fold_param.mix=mix;
  fold_param.midmix=midmix;
  fold_param.f_dim=f_dim;

  top=make_fold(&fold_param, levels,stop,(SIDE / pwidth));

  /* use first set of heights to set shadow value */
  shadow = extract(next_strip(top));
  a_strip = extract( next_strip(top) ); 
  b_strip = extract( next_strip(top) );

  /* initialise the light strengths */
  vstrength = vfract * contrast /( 1.0 + vfract );
  lstrength = contrast /( 1.0 + vfract );
}

#pragma mark -
#pragma mark === rendering ===
/*----------------------------------------------------------------------------------------
	° mirror
------------------------------------------------------------------------------------------*/

Col *MountainRenderer::mirror(
Height *a,
Height *b,
Height *shadow)
{
  Col *res, *map;
  Col last_col;
  int i,j, top, bottom, coord;
  int last_top, last_bottom;
  Height pivot;
  /* this routine returns a perspective view of the surface
   * with reflections in the water
   *
   */
  res = (Col *) malloc( s_height * sizeof(Col) );
  if( res == NULL )
  {
    fprintf(stderr,"malloc failed for picture strip\n");
    exit(1);
  }
  last_col=SKY;
  last_top=s_height-1;
  last_bottom=0;
  /*
   * many of the optimisation in the camera routine are
   * hard to implement in this case so we revert to the
   * simple painters algorithm modified to produce reflections
   * scan from back to front drawing strips between the
   * projected position of s_height and -s_height.
   * for water stipple the colour so the reflection is still visable
   */
  map=makemap(a,b,shadow);
  pivot=2.0*sealevel;
  for(i=width-1;i>0;i--)
  {
    if(map[i] < BAND_BASE)
    {
      /*{{{stipple water values*/
      for(j=last_bottom;j<=last_top;j++)
      {
        res[j]=last_col;
      }
      last_col=map[i];
      /* invalidate strip so last stip does not exist */
      last_bottom=s_height;
      last_top=-1;
      /* fill in water values */
      coord=1+project(i,sealevel);
      for(j=0;j<coord;j++)
      {
        /* do not print on every other point
         * if the current value is a land value
         */
        if( (j+base)%2 || (res[j]<BAND_BASE) )
        {
          res[j]=map[i];
        }
      }
      /* skip any adjacent bits of water with the same colour */
      while(map[i]==last_col)
      {
        i--;
      }
      i++;  /* the end of the for loop will decrement as well */
      /*}}}*/
    }else{
      /*{{{draw land values*/
      top = project(i,a[i]);
      bottom=project(i,pivot-a[i]);
      if(last_col == map[i])
      {
        if( top > last_top)
        {
          last_top = top;
        }
        if( bottom < last_bottom)
        {
          last_bottom=bottom;
        }
      }else{
        if(top < last_top)
        {
          for(j=top+1;j<=last_top;j++)
          {
            res[j]=last_col;
          }
        }
        if(bottom > last_bottom)
        {
          for(j=last_bottom;j<bottom;j++)
          {
            res[j]=last_col;
          }
        }
        last_top=top;
        last_bottom=bottom;
        last_col=map[i];
      }
      /*}}}*/
    }
  }
  /*{{{draw in front face*/
  for(j=last_bottom;j<=last_top;j++)
  {
    res[j]=last_col;
  }
  if( a[0] < sealevel )
  {
    coord=1+project(0,sealevel);
  }else{
    coord=1+project(0,a[0]);
  }
  for(j=0;j<coord;j++)
  {
    res[j] = map[0];
  }
  /*}}}*/
  base=1-base;
  free(map);
  return(res);
}
/*----------------------------------------------------------------------------------------
	° mirror
------------------------------------------------------------------------------------------*/

Col *MountainRenderer::camera(
Height *a,
Height *b,
Height *shadow)
{
  int i, j, coord, last;
  Col *res, col;

  /* this routine returns a perspective view of the surface */
  res = (Col *) malloc( s_height * sizeof(Col) );
  if( res == NULL )
  {
    fprintf(stderr,"malloc failed for picture strip\n");
    exit(1);
  }
  /*
   * optimised painters algorithm
   *
   * scan from front to back, we can avoid calculating the
   * colour if the point is not visable.
   */
  for( i=0, last=0 ; (i < width)&&(last < s_height) ; i++ )
  {
    if( a[i] < sealevel )
    {
      a[i] = sealevel;
    }
    coord = 1 + project( i, a[i] );
    if( coord > last )
    {
      /* get the colour of this point, the front strip should be black */
      if( i==0 )
      {
        col = BLACK;
      }else{
        col = get_col(b[i],a[i],b[i-1],shadow[i]);
      }
      if( coord > s_height )
      {
        coord = s_height;
      }
      for(;last<coord;last++)
      {
        res[last]=col;
      }
    }
  }
  for(;last<s_height;last++)
  {
    res[last]=SKY;
  }
  return(res);
}

/*----------------------------------------------------------------------------------------
	° mirror
------------------------------------------------------------------------------------------*/

Col *MountainRenderer::makemap (
Height *a,
Height *b,
Height *shadow)
{
Col *res;
int i;

  /* This routine returns a plan view of the surface */
  res = (Col *) malloc(width * sizeof(Col) );
  if (res == NULL)
  {
    fprintf(stderr,"malloc failed for colour strip\n");
    exit(1);
  }
  res[0] = BLACK;
  for(i=1 ; i<width ; i++)
  {
    res[i] = get_col(b[i],a[i],b[i-1],shadow[i]);
  }
  return(res);
}
/*----------------------------------------------------------------------------------------
	° mirror
------------------------------------------------------------------------------------------*/

Height *MountainRenderer::extract (
Strip *s)
{
  int i;
  
  Height *p;
  p = s->d;
  free(s);
  for(i=0 ; i<width; i++ )
  {
    p[i] = shift + (vscale * p[i]);
  }
  return(p);
}

/*----------------------------------------------------------------------------------------
	° project
------------------------------------------------------------------------------------------*/

int MountainRenderer::project (
int x,
Height y)
{
  int pos;
#ifndef SLOPPY
  double theta;

  theta = atan( (double) ((viewheight - y)/( x - viewpos)) );
  theta = theta - vangle;
  pos = (s_height/2) - (focal * tan( theta));
#else
  float tan_theta;

  /* nast approx to avoid trig functions */
  tan_theta = (viewheight -y)/(x-viewpos) - tan_vangle;
  pos = (s_height/2) - (focal * tan_theta);
#endif
  if( pos > (s_height-1))
  {
    pos = s_height-1;
  }
  else if( pos < 0 )
  {
    pos = 0;
  }
  return( pos );
}

/*----------------------------------------------------------------------------------------
	° get_col
------------------------------------------------------------------------------------------*/

Col MountainRenderer::get_col (
Height p,
Height p_minus_x,
Height p_minus_y,
Height shadow)
{
  Height delta_x, delta_y;
  Height delta_x_sqr, delta_y_sqr;
  Height hypot_sqr;
  
  double norm, dshade;
  Height effective;
  Col result;
  int band, shade;
  /*{{{  if underwater*/
  if ( p < sealevel )
  {
    if( shadow > sealevel )
    {
      return( SEA_UNLIT );
    }else{
      return( SEA_LIT );
    }
  }
  /*}}}*/
  /*
   * We have three light sources, one slanting in from the left
   * one directly from above and an ambient light.
   * For the directional sources illumination is proportional to the
   * cosine between the normal to the surface and the light.
   *
   * The surface contains two vectors
   * ( 1, 0, delta_x )
   * ( 0, 1, delta_y )
   *
   * The normal therefore is parallel to
   * (  -delta_x, -delta_y, 1)/sqrt( 1 + delta_x^2 + delta_y^2)
   *
   * For light parallel to ( cos_phi, 0, -sin_phi) the cosine is
   *        (cos_phi*delta_x + sin_phi)/sqrt( 1 + delta_x^2 + delta_y^2)
   *
   * For light parallel to ( cos_phi*cos_alpha, cos_phi*sin_alpha, -sin_phi)
   * the cosine is
   * (cos_phi*cos_alpha*delta_x + cos_phi*sin_alpha*delta_y+ sin_phi)/sqrt( 1 + delta_x^2 + delta_y^2)
   *
   * For vertical light the cosine is
   *        1 / sqrt( 1 + delta_x^2 + delta_y^2)
   */
   
  delta_x = p - p_minus_x;
  delta_y = p - p_minus_y;
  delta_x_sqr = delta_x * delta_x;
  delta_y_sqr = delta_y * delta_y;
  hypot_sqr = delta_x_sqr + delta_y_sqr;
  norm = sqrt( 1.0 + hypot_sqr );

  /*{{{  calculate effective s_height */
  effective = (p + (varience * contour *
          (1.0/ ( 1.0 + hypot_sqr))));
  /*}}}*/
  /*{{{  calculate colour band. */
  band = ( effective / varience) * N_BANDS;
  if ( band < 0 )
  {
    band = 0;
  }
  if( band > (N_BANDS - 1))
  {
    band = (N_BANDS -1);
  }
  result = (BAND_BASE + (band * band_size));
  /*}}}*/

  /*{{{calculate the illumination stength*/
  /*
   * add in a contribution for the vertical light. The normalisation factor
   * is applied later
   *
   */
  dshade = vstrength;
  
  if( p >= shadow )
  {
    /*
     * add in contribution from the main light source
     */
    /* dshade += ((double) lstrength * ((delta_x * cos_phi) + sin_phi));*/
    dshade += ((double) lstrength *
               ((delta_x * x_fact) + (delta_y * y_fact) + sin_phi));
  }
  /* divide by the normalisation factor (the same for both light sources) */
  dshade /= norm;
  /*}}}*/
  /*{{{  calculate shading */
  /* dshade should be in the range 0.0 -> 1.0
   * if the light intensities add to 1.0
   * now convert to an integer
   */
  shade = dshade * (double) band_size;
  if( shade > (band_size-1))
  {
    shade = (band_size-1);
  }
  /*{{{  if shade is negative then point is really in deep shadow */
  if( shade < 0 )
  {
      shade = 0;
  }
  /*}}}*/
  /*}}}*/
  result += shade;
  if( (result >= n_col) || (result < 0) )
  {
    fprintf(stderr,"INTERNAL ERROR colour out of range %d\n",result);
    exit(1);
  }
  return(result);
}

