#include "MessageUtils.h"

#include "ConfigView.h"
#include "MountainModule.h"

char *app_signature = "application/x-vnd.rickb-mountainmodule";

// MAIN INSTANTIATION FUNCTION
BArchivable *BlanketModule::Instantiate(BMessage* message)
{
	return new MountainModule(message);
}

#pragma mark -
#pragma mark == MountainModule ===
/*----------------------------------------------------------------------------------------
	° MountainModule
------------------------------------------------------------------------------------------*/

MountainModule::MountainModule(BMessage* archive)
 : BlanketModule(archive)
{
	mRenderer = new MountainRenderer;
	int32 size;
	const void * s = NULL;
	archive->FindData("mRenderer",B_RAW_TYPE,&s,&size);
	if(s)
		memcpy(mRenderer, s, sizeof(MountainRenderer));
	mRenderer->init_graphics();
}

MountainModule::~MountainModule()
{
	delete mRenderer;
}

/*----------------------------------------------------------------------------------------
	° Archive
------------------------------------------------------------------------------------------*/

status_t MountainModule::Archive(BMessage *archive, bool deep) const
{
	BlanketModule::Archive(archive,deep);

	archive->AddString("class","MountainModule");
	archive->AddData("mRenderer",B_RAW_TYPE,mRenderer,sizeof(MountainRenderer));

	return B_OK;
}

/*----------------------------------------------------------------------------------------
	° InstantiateSaver
------------------------------------------------------------------------------------------*/

BlanketModule::Saver* MountainModule::InstantiateSaver(BView *v, bool /*preview*/)
{
	return new MountainSaver(this, v);
}

/*----------------------------------------------------------------------------------------
	° InstantiateConfig
------------------------------------------------------------------------------------------*/

BlanketModule::Config* MountainModule::InstantiateConfig(BView *v)
{
	return new MountainConfig(this, v);
}	
#pragma mark -
#pragma mark == MountainSaver ===
/*----------------------------------------------------------------------------------------
	° MountainSaver
------------------------------------------------------------------------------------------*/

MountainSaver::MountainSaver(BlanketModule *owner, BView* v)
 : BlanketModule::Saver(owner, v, 0)
{
	mRenderer = ((MountainModule*)owner)->mRenderer;
	BRect rect(0,0,mRenderer->s_width - 1, mRenderer->s_height - 1);
	mImage = new BBitmap(rect,B_COLOR_8_BIT,true);
	mImage->AddChild(new BView(rect,"",0,0));
}

MountainSaver::~MountainSaver()
{
	delete mImage;
}

/*----------------------------------------------------------------------------------------
	° Tick
------------------------------------------------------------------------------------------*/

// This is where all the real work gets done to draw into the view
void MountainSaver::Tick()
{
	mRenderer->PlotNextColumn(mImage);	
	
	BAutolock lock(View()->Window());		
	BRect bounds = View()->Bounds();
	BRect src = bounds;
	
	if(!mRenderer->mScrolling)
	{	
		src.top = 0; src.bottom = mRenderer->s_height;
		// rather than sync drawing, we'll just invalidate a bigger rect.
		src.left  = mRenderer->mCurCol - 5;
		src.right = mRenderer->mCurCol;
	}
	
	BRect dst = src;
	dst.OffsetBy((bounds.Width()  - mRenderer->s_width )/2,
				 (bounds.Height() - mRenderer->s_height)/2);
	
	View()->DrawBitmapAsync(mImage,src,dst);
}

#pragma mark -
#pragma mark == MountainConfig ===
/*----------------------------------------------------------------------------------------
	° MountainConfig
------------------------------------------------------------------------------------------*/

MountainConfig::MountainConfig(BlanketModule *owner, BView* view)
 : BlanketModule::Config(owner, view, false)
{
	BMimeType mime(app_signature);
	if(!mime.IsInstalled())
		mime.Install();

	entry_ref ref;
	be_roster->FindApp(app_signature,&ref);
	BFile file(&ref, B_READ_ONLY);
	BResources res;
	status_t status = res.SetTo(&file);
	
	// add the config view
	if(status == B_OK)
	{
		MountainConfigView* config_view = cast_as(MessageUtils::RestoreFromResource(&res, 'ARCV', 1), MountainConfigView);
		if(config_view != NULL)
		{	
			view->AddChild(config_view);
			BAutolock lock(view->Window());
			config_view->SyncControls(((MountainModule*)Owner())->mRenderer);	
		}else
			status = B_ERROR;
	}
	if(status != B_OK){
		view->AddChild(new BStringView(BRect(10, 10, 200, 35), B_EMPTY_STRING, "Error Loading Resources."));	
		view->AddChild(new BStringView(BRect(10, 35, 200, 60), B_EMPTY_STRING, "Try running mimeset on MountainModule."));	
	}
}

MountainConfig::~MountainConfig()
{
	BAutolock lock(View()->Window());
	MountainConfigView* config_view = cast_as(View()->FindView("ConfigView"),MountainConfigView);
	if(config_view != NULL)
		config_view->Settings(((MountainModule*)Owner())->mRenderer);
}